/* iBurst (TM) compatible driver for 2.6 Linux kernel.
 * based on the original ArrayComm (TM) iBurst (TM) driver.
 * Nicholas Jefferson <nicholas@pythontraining.com.au>
 * 11 May 2005
 *
 * Fixes by Nik Trevallyn-Jones.
 * Fixes to support new hardware by Shane MacPhillamy.
 * Support for 2.6.20 by Daniel Burr.
 *
 * Wireless Extension IOCTL code based on that in
 * orinoco.c (v0.15rc2 28 July 2004) in the orinoco driver, written by
 * Pavel Roskin, David Gibson, Jean Tourrilhes & Benjamin Herrenschmidt.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "ib-net.h"

#include <linux/version.h>
#include <linux/etherdevice.h>
#include <net/iw_handler.h>

#define IB_NET_QUEUE_LO 4
#define IB_NET_QUEUE_HI 10
#define IB_PROTO_CONTROL 0x00
#define IB_CONTROL_STATUS1 0x05

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
#define IW_QUAL_QUAL_UPDATED	0x01	/* Value was updated since last read */
#define IW_QUAL_LEVEL_UPDATED	0x02
#define IW_QUAL_NOISE_UPDATED	0x04
#endif

enum ib_net_offset
{
	OFFSET_STATUS1_NPAYLOAD = 0x00,
	OFFSET_STATUS1_TYPE = 0x01,
	OFFSET_STATUS1_LEVEL = 0x02,
	OFFSET_STATUS1_UPCOUNT = 0x03,
	OFFSET_STATUS1_DOWNCOUNT = 0x04,
	OFFSET_STATUS1_SAMPLE = 0x05,
	OFFSET_STATUS1_SUCCESS = 0x0a,
	OFFSET_STATUS1_ATTEMPT = 0x0b,
	OFFSET_STATUS1_SINR = 0x0c,
};

static int debug = 0;
module_param(debug, int, 0);
#define DEBUG(n, args...) if (debug < (n)) ; else printk(KERN_INFO args)

/**
 * ifname - interface name.
 */
static char *ifname = "ib%d";
module_param(ifname, charp, 0);

spinlock_t ib_lock = SPIN_LOCK_UNLOCKED;

/**
 * ib_net_rx_deliver - deliver received radio frame as ethernet frame.
 *     called under ib_lock
 * @modem: modem state.
 * @nbuf: radio frame size.
 */
static void ib_net_rx_deliver(struct ib_net_modem_t *modem, int nbuf)
{
	struct ib_net_radio_t *radio = (struct ib_net_radio_t*) modem->rx_buf;
	struct ib_net_ether_t *ether;
	struct sk_buff *skb;
	int npayload;

	modem->stats.rx_bytes += nbuf;
	modem->stats.rx_packets += 1;
	modem->netdev->last_rx = jiffies;
	npayload = nbuf - IB_NET_RADIO_HEAD;
	skb = alloc_skb(IB_NET_ETHER_HEAD + npayload, GFP_ATOMIC);
	if (skb == NULL) {
		modem->stats.rx_dropped += 1;
		return;
	}
	skb->dev = modem->netdev;
	ether = (struct ib_net_ether_t*)
			skb_put(skb, IB_NET_ETHER_HEAD + npayload);
	memcpy(ether->source, modem->netdev->dev_addr, ETH_ALEN);
	if (radio->word[0] & FLAG_BROADCAST)
		memset(ether->dest, 0xff, ETH_ALEN);
	else {
		ether->source[ETH_ALEN - 1] ^= 1;
		memcpy(ether->dest, modem->netdev->dev_addr, ETH_ALEN);
	}
	ether->proto[0] = radio->proto[0];
	ether->proto[1] = radio->proto[1];
	memcpy(ether->payload, radio->payload, npayload);
	skb->protocol = eth_type_trans(skb, modem->netdev);
	skb->ip_summed = CHECKSUM_UNNECESSARY;
	netif_rx(skb);
}

/**
 * ib_net_rx_control_parse - process received control frame.
 *     called under ib_lock
 * @modem: modem state.
 * @nbuf: control frame size.
 */
static void ib_net_rx_control_parse(struct ib_net_modem_t *modem, int nbuf)
{
	struct ib_net_radio_t *radio = (struct ib_net_radio_t*) modem->rx_buf;
	int32_t *payload = (int32_t*) &radio->payload;
	int npayload, type, level, sinr, attempt, success;
#if 17 <= WIRELESS_EXT
	int sample, upcount, downcount;
#endif

	level = 0;	// to silence a benign warning
	npayload = nbuf - IB_NET_RADIO_HEAD;
	if (npayload < 8)
		return;
	if (npayload != be32_to_cpup(payload + OFFSET_STATUS1_NPAYLOAD))
		return;
	type = be32_to_cpup(payload + OFFSET_STATUS1_TYPE);
	if (type == IB_CONTROL_STATUS1 && 12 <= npayload) {
		/**
		 * level is reported as an int 0 <= n <= 100,
		 * or -1 => invalid
		 */
		level = le32_to_cpup(payload + OFFSET_STATUS1_LEVEL);
		if (level != -1) {
			modem->wstats.qual.level = modem->wstats.qual.qual
					= level;
			modem->wstats.qual.updated |= IW_QUAL_LEVEL_UPDATED
					| IW_QUAL_QUAL_UPDATED;
		}
	}
#if 17 <= WIRELESS_EXT
	if (type == IB_CONTROL_STATUS1 && 24 <= npayload) {
		/**
		 * up and down throughput are reported as int byte
		 * counts; the sample time is reported in milliseconds.
		 *
		 * The rates are calculated as bits/sec.
		 */
		sample = le32_to_cpup(payload + OFFSET_STATUS1_SAMPLE);
		upcount = le32_to_cpup(payload + OFFSET_STATUS1_UPCOUNT);
		downcount = le32_to_cpup(payload + OFFSET_STATUS1_DOWNCOUNT);
		if (sample) {
			modem->uprate = upcount * 1000 / sample * 8;
			modem->downrate = downcount * 1000 / sample * 8;
		}
	}
#endif
	if (type == IB_CONTROL_STATUS1 && 48 <= npayload) {
		/**
		 * success and attempt report the number of bursts
		 * attempted/successful since last report.
		 *
		 * Missed bursts are calculated as attempt - success.
		 */
		attempt = le32_to_cpup(payload + OFFSET_STATUS1_ATTEMPT);
		success = le32_to_cpup(payload + OFFSET_STATUS1_SUCCESS);
		modem->wstats.miss.beacon = attempt - success;
	}
	if (type == IB_CONTROL_STATUS1 && 52 <= npayload) {
		/**
		 * sinr (signal to interference AND noise) is reported as
		 * actual-sinr << 4 (ie sinr * 16).
		 *
		 * noise is calculated as level / (sinr+1).
		 *
		 * Since sinr is already multiplied by 16, top and bottom of
		 * ratio are multiplied by 16. Hence level => level*16 and
		 * sinr+1 => sinr+16 (sinr is already multiplied by 16).
		 */
		sinr = le32_to_cpup(payload + OFFSET_STATUS1_SINR);
		if (sinr) {
			modem->wstats.qual.noise = (level * 16) / (sinr + 16);
			modem->wstats.qual.updated |= IW_QUAL_NOISE_UPDATED;
		}
	}
}

/**
 * ib_net_rx_parse - process received radio frame.
 *     called under ib_lock
 * @modem: modem state.
 * @nbuf: radio frame size.
 */
void ib_net_rx_parse(struct ib_net_modem_t *modem, int nbuf)
{
	struct ib_net_radio_t *radio = (struct ib_net_radio_t*) modem->rx_buf;

	if (radio->word[0] & FLAG_EXTENSION)
		return;
	if (radio->word[1] ^ radio->check ^ 0xff)
		return;
	if (radio->proto[0] != IB_PROTO)
		ib_net_rx_deliver(modem, nbuf);
	else if (radio->proto[1] == IB_PROTO_CONTROL)
		ib_net_rx_control_parse(modem, nbuf);
	else if (modem->driver->rx_parse)
		modem->driver->rx_parse(modem->pdriver);
}

/**
 * ib_net_tx_prepare - prepare next radio frame for transmission.
 *     called under ib_lock
 * @modem: modem state.
 */
int ib_net_tx_prepare(struct ib_net_modem_t *modem)
{
	struct sk_buff *skb;
	struct ib_net_radio_t *radio;
	int nqueue;

	nqueue = skb_queue_len(&modem->tx_queue);
	if (nqueue < IB_NET_QUEUE_LO)
		netif_wake_queue(modem->netdev);
	skb = skb_dequeue(&modem->tx_queue);
	if (skb == NULL)
		return 0;
	memcpy(modem->tx_buf, skb->data, skb->len);
	dev_kfree_skb(skb);
	radio = (struct ib_net_radio_t*) modem->tx_buf;
	radio->packet = modem->stats.tx_packets & 0xff;
	modem->netdev->trans_start = jiffies;
	return skb->len;
}

/**
 * ib_net_deregister - deregister network device.
 * @modem: modem state.
 */
void ib_net_deregister(struct ib_net_modem_t *modem)
{
	unregister_netdev(modem->netdev);
	free_netdev(modem->netdev);
}

/**
 * ib_net_poll - bottom half handler for interrupts.
 * @_modem: modem state.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
static void ib_net_poll(void *_modem)
#else
static void ib_net_poll(struct work_struct *work)
#endif
{
	unsigned long state;
	struct ib_net_modem_t *modem = 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
		_modem;
#else
		container_of(work, struct ib_net_modem_t, work_queue);
#endif

	spin_lock_irqsave(&ib_lock, state);
	if (modem->driver->poll)
		modem->driver->poll(modem->pdriver);
	spin_unlock_irqrestore(&ib_lock, state);
}

/**
 * ib_net_stats - return network device statistics.
 * @netdev: network device state.
 */
static struct net_device_stats *ib_net_stats(struct net_device *netdev)
{
	struct ib_net_modem_t *modem = netdev->priv;

	return &modem->stats;
}

/**
 * ib_wireless_stats - return wireless statistics.
 * @netdev: network device state.
 */
static struct iw_statistics *ib_wireless_stats(struct net_device *netdev)
{
	struct ib_net_modem_t *modem = netdev->priv;

	return &modem->wstats;
}

/**
 * ib_net_tx_timeout - timeout handler.
 * @netdev: network device state.
 */
static void ib_net_tx_timeout(struct net_device *netdev)
{
	unsigned long state;
	struct ib_net_modem_t *modem = netdev->priv;

	spin_lock_irqsave(&ib_lock, state);
	if (modem->driver->tx_timeout)
		modem->driver->tx_timeout(modem->pdriver);
	netdev->trans_start = jiffies;
	spin_unlock_irqrestore(&ib_lock, state);
}

/**
 * ib_net_tx_start - convert ethernet frame to radio frame for transmission.
 * @skb: ethernet frame.
 * @netdev: network device state.
 */
static int ib_net_tx_start(struct sk_buff *skb, struct net_device *netdev)
{
	unsigned long state;
	struct ib_net_modem_t *modem = netdev->priv;
	struct ib_net_ether_t *ether;
	struct ib_net_radio_t *radio;
	int nbuf, nqueue;

	ether = (struct ib_net_ether_t*) skb->data;
	radio = (struct ib_net_radio_t*)
			skb_pull(skb, IB_NET_ETHER_HEAD - IB_NET_RADIO_HEAD);
	nbuf = skb->len;
	if (ether->dest[0] & 1)
		radio->word[0] = (nbuf >> 8) | FLAG_BROADCAST;
	else
		radio->word[0] = nbuf >> 8;
	radio->word[1] = nbuf & 0xff;
	radio->check = (nbuf & 0xff) ^ 0xff;
	spin_lock_irqsave(&ib_lock, state);
	skb_queue_tail(&modem->tx_queue, skb);
	nqueue = skb_queue_len(&modem->tx_queue);
	if (IB_NET_QUEUE_HI < nqueue)
		netif_stop_queue(modem->netdev);
	if (modem->driver->poll)
		modem->driver->poll(modem->pdriver);
	spin_unlock_irqrestore(&ib_lock, state);
	return 0;
}

/**
 * ib_net_open - open network device.
 * @netdev: network device state.
 */
static int ib_net_open(struct net_device *netdev)
{
	struct ib_net_modem_t *modem = netdev->priv;
	int err;

	spin_lock_irq(&ib_lock);
	if (modem->driver->open) {
		err = modem->driver->open(modem->pdriver);
		if (!err)
			netif_start_queue(netdev);
	} else
		err = -ENODEV;
	spin_unlock_irq(&ib_lock);
	return err;
}

/**
 * ib_net_close - close network device.
 * @netdev: network device state.
 */
static int ib_net_close(struct net_device *netdev)
{
	struct ib_net_modem_t *modem = netdev->priv;
	int err;

	spin_lock_irq(&ib_lock);
	if (modem->driver->close) {
		err = modem->driver->close(modem->pdriver);
		if (!err) {
			modem->wstats.qual.level = 0;
			modem->wstats.qual.qual = 0;
			modem->wstats.qual.noise = 0;
			modem->wstats.qual.updated = 0x7;
			netif_stop_queue(netdev);
		}
	} else
		err = -ENODEV;
	spin_unlock_irq(&ib_lock);
	return err;
}

/**
 * ib_net_change_mtu - change mtu.
 * @netdev: network device state.
 * @mtu: new mtu.
 */
static int ib_net_change_mtu(struct net_device *netdev, int mtu)
{
	if (mtu < 0 || ETH_DATA_LEN < mtu)
		return -EINVAL;
	netdev->mtu = mtu;
	return 0;
}

#if 17 <= WIRELESS_EXT
/**
 * ib_ioctl_getname - return the name of this wireless interface
 *
 * The header file suggests (strongly) that this indicate the wireless
 * protocol (eg "IEEE 802.11").
 *
 * @dev: network device state (ignored).
 * @info: request info (ignored).
 * @name: interface name (returned).
 * @extra: extra flags (ignored).
 */
static int ib_ioctl_getname(struct net_device *dev,
		struct iw_request_info *info, char *name, char *extra)
{
	strcpy(name, "HC-SDMA (iBurst)");
	return 0;
}

/**
 * ib_ioctl_getrate - return the current date rate of this wireless interface
 *
 * Wireless Extensions make no distinction between up and down rates,
 * so we report the larger of the two. One way of looking at this is that
 * we report on whatever is doing the most work. Another way of looking
 * at it is that we report the most impressive number...
 *
 * @dev: network device state.
 * @info: request info (ignored).
 * @rrq: request data (returned)
 * @extra: extra flags (ignored).
 */
static int ib_ioctl_getrate(struct net_device *dev,
		struct iw_request_info *info, struct iw_param *rrq, char *extra)
{
	struct ib_net_modem_t *modem = dev->priv;

	rrq->value = (modem->downrate <= modem->uprate
			? modem->uprate : modem->downrate);
	rrq->fixed = 0;
	rrq->disabled = 0;
	return 0;
}

/**
 * ib_ioctl_getrange - return the range information about other values.
 *
 * We leave the majority of the iw_range struct unfilled-in, since the
 * device does not support/report values for most fields.
 *
 * @dev: network device state (ignored).
 * @info: request info (ignored).
 * @rrq: request data (returned)
 * @extra: the iw_range struct (returned).
 */
static int ib_ioctl_getrange(struct net_device *dev,
		struct iw_request_info *info, struct iw_point *rrq, char *extra)
{
	struct iw_range *range = (struct iw_range*) extra;

	rrq->length = sizeof(struct iw_range);
	memset(range, 0, sizeof(struct iw_range));
	range->we_version_compiled = WIRELESS_EXT;
	range->max_qual.qual = 100;
	range->max_qual.level = 100;
	range->max_qual.noise = 100;
	range->num_bitrates = 1;
	range->throughput = 3000000;
	return 0;
}

/**
 * ib_wireless_handler - structure to export the Wireless Handlers.
 *
 * This is an array of iw_handler's (wireless ioctl functions).
 */
static const iw_handler ib_wireless_handler[] =
{
	[SIOCGIWNAME - SIOCIWFIRST] = (iw_handler) ib_ioctl_getname,
	[SIOCGIWRANGE - SIOCIWFIRST] = (iw_handler) ib_ioctl_getrange,
	[SIOCGIWRATE - SIOCIWFIRST] = (iw_handler) ib_ioctl_getrate,
};

/**
 * ib_wireless_def - wireless functions.
 */
static struct iw_handler_def ib_wireless_def = {
	.num_standard = sizeof(ib_wireless_handler) / sizeof(iw_handler),
	.standard = (iw_handler*) ib_wireless_handler,
	.get_wireless_stats = ib_wireless_stats,
};
#endif

/**
 * ib_net_setup - initialize net_device structure.
 * @netdev: network device state.
 */
static void ib_net_setup(struct net_device *netdev)
{
	ether_setup(netdev);
	netdev->flags = IFF_BROADCAST | IFF_DYNAMIC | IFF_NOARP;
	netdev->get_stats = ib_net_stats;
#if WIRELESS_EXT < 17
	netdev->get_wireless_stats = ib_wireless_stats;
#else
	netdev->wireless_handlers = &ib_wireless_def;
#endif
	netdev->hard_start_xmit = ib_net_tx_start;
	netdev->open = ib_net_open;
	netdev->stop = ib_net_close;
	netdev->trans_start = jiffies;
	netdev->tx_timeout = ib_net_tx_timeout;
	netdev->watchdog_timeo = 10 * HZ;
	netdev->change_mtu = ib_net_change_mtu;
	SET_MODULE_OWNER(netdev);
}

/**
 * ib_net_register - register network device.
 * @_netdev: return network device state.
 */
int ib_net_register(struct net_device **_netdev)
{
	struct net_device *netdev;
	struct ib_net_modem_t *modem;
	int err;

	netdev = alloc_netdev(sizeof(struct ib_net_modem_t), ifname,
			ib_net_setup);
	if (netdev == NULL)
		return -ENOMEM;
	modem = netdev->priv;
	modem->netdev = netdev;
	modem->pc_status = 0;
	modem->ut_status = 0;
	memset(modem->tx_buf, 0, sizeof(modem->tx_buf));
	memset(&modem->stats, 0, sizeof(struct net_device_stats));
	memset(&modem->wstats, 0, sizeof(struct iw_statistics));
	err = register_netdev(netdev);
	if (err) {
		DEBUG(1, "ib-net: register_netdev failed\n");
		free_netdev(netdev);
		return err;
	}
	skb_queue_head_init(&modem->tx_queue);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
	INIT_WORK(&modem->work_queue, ib_net_poll, modem);
#else
	INIT_WORK(&modem->work_queue, ib_net_poll);
#endif
	*_netdev = netdev;
	return 0;
}

EXPORT_SYMBOL(ib_net_register);
EXPORT_SYMBOL(ib_net_deregister);
EXPORT_SYMBOL(ib_net_rx_parse);
EXPORT_SYMBOL(ib_net_tx_prepare);
EXPORT_SYMBOL(ib_lock);

MODULE_DESCRIPTION("iBurst compatible driver");
MODULE_LICENSE("GPL");
