/*
 *	Device handling code
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek		<buytenh@gnu.org>
 *
 *	$Id: br_device.c,v 1.3 2008/04/08 11:27:03 qjj_qin Exp $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/ethtool.h>

#include <asm/uaccess.h>
#include "br_private.h"

#ifdef IGMP_SNOOPING
/*2008-01-15,for porting igmp snooping to linux kernel 2.6*/
#include <linux/ip.h>
#include <linux/in.h>
#include <linux/igmp.h>
#include <net/checksum.h>
#ifdef CONFIG_BRIDGE_IGMPV3_SNOOPING
#include "../../drivers/net/re865x/igmpsnooping/rtl865x_igmpsnooping_glue.h"
#include "../../drivers/net/re865x/igmpsnooping/rtl865x_igmpsnooping.h"
#ifdef CONFIG_RTL865X_HARDWARE_MULTICAST
#include "../../drivers/net/re865x/l3Driver/rtl865x_multicast.h"
#endif
extern unsigned int brIgmpModuleIndex;
#endif
#endif

static struct net_device_stats *br_dev_get_stats(struct net_device *dev)
{
	struct net_bridge *br = netdev_priv(dev);
	return &br->statistics;
}

/* net device transmit always called with no BH (preempt_disabled) */
int br_dev_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct net_bridge *br = netdev_priv(dev);
	const unsigned char *dest = skb->data;
	struct net_bridge_fdb_entry *dst;
#ifdef IGMP_SNOOPING	
/*2008-01-15,for porting igmp snooping to linux kernel 2.6*/
	struct iphdr *iph;
	unsigned int ipaddr=0;
	unsigned char proto=0;
	unsigned char reserved=0;
	extern int igmpsnoopenabled;
	
	#ifdef CONFIG_BRIDGE_IGMPV3_SNOOPING
	struct rtl_multicastDataInfo multicastDataInfo;
	#endif
	
	iph = skb->nh.iph;
	ipaddr =  iph->daddr;
	proto =  iph->protocol;
	 if((ipaddr&0xFFFFFF00)==0xE0000000)
	        reserved=1;
	 
	#ifdef CONFIG_BRIDGE_IGMPV3_SNOOPING
	multicastDataInfo.ipVersion=4;
	multicastDataInfo.sourceIp[0]=  (uint32)(iph->saddr);
	multicastDataInfo.groupAddr[0]=  (uint32)(iph->daddr);
	#ifdef CONFIG_BRIDGE_IGMPV3_SNOOPING_DEBUG
	printk("source ip is 0x%x, dest ip is 0x%x\n",multicastDataInfo.sourceIp[0],multicastDataInfo.groupAddr[0]);
	#endif
	#endif
	
#endif	

	br->statistics.tx_packets++;
	br->statistics.tx_bytes += skb->len;

	skb->mac.raw = skb->data;
	skb_pull(skb, ETH_HLEN);

#ifdef IGMP_SNOOPING
/*2008-01-15,for porting igmp snooping to linux kernel 2.6*/
	if (dest[0] & 1)
	{
		if(igmpsnoopenabled && MULTICAST_MAC(dest) && (proto !=IPPROTO_IGMP) && (ipaddr != 0xEFFFFFFA) && (reserved ==0)) 
		{	
			#ifdef CONFIG_BRIDGE_IGMPV3_SNOOPING
			int retVal=FAILED;
			struct rtl_multicastFwdInfo multicastFwdInfo;
			#ifdef CONFIG_RTL865X_HARDWARE_MULTICAST
			unsigned int srcPort=skb->srcPort;
			unsigned int srcVlanId=skb->srcVlanId;
			#endif
			retVal= rtl_getMulticastDataFwdInfo(brIgmpModuleIndex, &multicastDataInfo, &multicastFwdInfo);
			if(retVal==SUCCESS)
			{
			
				br_multicast_deliver(br, multicastFwdInfo.fwdPortMask, skb, 0);

				#ifdef CONFIG_RTL865X_HARDWARE_MULTICAST
				if((srcVlanId!=0) && (srcPort!=0xFFFF))
				{
					#if defined(CONFIG_RTK_VLAN_SUPPORT)
					if(rtk_vlan_support_enable == 0)
					{
						rtl865x_ipMulticastHardwareAccelerate(br, multicastFwdInfo.fwdPortMask,srcPort,srcVlanId, multicastDataInfo.sourceIp[0], multicastDataInfo.groupAddr[0]);
					}
					#else
					rtl865x_ipMulticastHardwareAccelerate(br, multicastFwdInfo.fwdPortMask,srcPort,srcVlanId, multicastDataInfo.sourceIp[0], multicastDataInfo.groupAddr[0]);
					#endif
				}
				#endif		
			}
			else
			{
				br_flood_deliver(br, skb, 0);
			}
			#else
			if ((dst = br_fdb_get(br, dest)) != NULL)
			{
				br_multicast_deliver(br, dst, skb, 0);
				br_fdb_put(dst);
			}
			else
			{
				br_flood_deliver(br, skb, 0);
			}
			#endif
		}
		else
		{ 
			// broadcast
			br_flood_deliver(br, skb, 0);
		}	
	
	}
	else if ((dst = __br_fdb_get(br, dest)) != NULL)
	{
		br_deliver(dst->dst, skb);
	}
	else
	{
		br_flood_deliver(br, skb, 0);
	}
#else
	if (dest[0] & 1) 
		br_flood_deliver(br, skb, 0);
	else if ((dst = __br_fdb_get(br, dest)) != NULL)
		br_deliver(dst->dst, skb);
	else
		br_flood_deliver(br, skb, 0);
#endif

	return 0;
}

static int br_dev_open(struct net_device *dev)
{
	struct net_bridge *br = netdev_priv(dev);

	br_features_recompute(br);
	netif_start_queue(dev);
	br_stp_enable_bridge(br);

	return 0;
}

static void br_dev_set_multicast_list(struct net_device *dev)
{
}

static int br_dev_stop(struct net_device *dev)
{
	br_stp_disable_bridge(netdev_priv(dev));

	netif_stop_queue(dev);

	return 0;
}

static int br_change_mtu(struct net_device *dev, int new_mtu)
{
	if (new_mtu < 68 || new_mtu > br_min_mtu(netdev_priv(dev)))
		return -EINVAL;

	dev->mtu = new_mtu;
	return 0;
}

/* Allow setting mac address of pseudo-bridge to be same as
 * any of the bound interfaces
 */
static int br_set_mac_address(struct net_device *dev, void *p)
{
	struct net_bridge *br = netdev_priv(dev);
	struct sockaddr *addr = p;
	struct net_bridge_port *port;
	int err = -EADDRNOTAVAIL;

	spin_lock_bh(&br->lock);
	list_for_each_entry(port, &br->port_list, list) {
		if (!compare_ether_addr(port->dev->dev_addr, addr->sa_data)) {
			br_stp_change_bridge_id(br, addr->sa_data);
			err = 0;
			break;
		}
	}
	spin_unlock_bh(&br->lock);

	return err;
}

static void br_getinfo(struct net_device *dev, struct ethtool_drvinfo *info)
{
	strcpy(info->driver, "bridge");
	strcpy(info->version, BR_VERSION);
	strcpy(info->fw_version, "N/A");
	strcpy(info->bus_info, "N/A");
}

static int br_set_sg(struct net_device *dev, u32 data)
{
	struct net_bridge *br = netdev_priv(dev);

	if (data)
		br->feature_mask |= NETIF_F_SG;
	else
		br->feature_mask &= ~NETIF_F_SG;

	br_features_recompute(br);
	return 0;
}

static int br_set_tso(struct net_device *dev, u32 data)
{
	struct net_bridge *br = netdev_priv(dev);

	if (data)
		br->feature_mask |= NETIF_F_TSO;
	else
		br->feature_mask &= ~NETIF_F_TSO;

	br_features_recompute(br);
	return 0;
}

static int br_set_tx_csum(struct net_device *dev, u32 data)
{
	struct net_bridge *br = netdev_priv(dev);

	if (data)
		br->feature_mask |= NETIF_F_NO_CSUM;
	else
		br->feature_mask &= ~NETIF_F_ALL_CSUM;

	br_features_recompute(br);
	return 0;
}

static struct ethtool_ops br_ethtool_ops = {
	.get_drvinfo = br_getinfo,
	.get_link = ethtool_op_get_link,
	.get_sg = ethtool_op_get_sg,
	.set_sg = br_set_sg,
	.get_tx_csum = ethtool_op_get_tx_csum,
	.set_tx_csum = br_set_tx_csum,
	.get_tso = ethtool_op_get_tso,
	.set_tso = br_set_tso,
};

void br_dev_setup(struct net_device *dev)
{
	memset(dev->dev_addr, 0, ETH_ALEN);

	ether_setup(dev);

	dev->do_ioctl = br_dev_ioctl;
	dev->get_stats = br_dev_get_stats;
	dev->hard_start_xmit = br_dev_xmit;
	dev->open = br_dev_open;
	dev->set_multicast_list = br_dev_set_multicast_list;
	dev->change_mtu = br_change_mtu;
	dev->destructor = free_netdev;
	SET_MODULE_OWNER(dev);
 	SET_ETHTOOL_OPS(dev, &br_ethtool_ops);
	dev->stop = br_dev_stop;
	dev->tx_queue_len = 0;
	dev->set_mac_address = br_set_mac_address;
	dev->priv_flags = IFF_EBRIDGE;

 	dev->features = NETIF_F_SG | NETIF_F_FRAGLIST | NETIF_F_HIGHDMA |
 			NETIF_F_TSO | NETIF_F_NO_CSUM | NETIF_F_GSO_ROBUST;
}


#ifdef CONFIG_RTL865X_HARDWARE_MULTICAST



int rtl865x_genMCastFwdDescriptor(char *devName, uint32 devPortNum,struct rtl_multicastDataInfo *multicastDataInfoPtr, rtl865x_mcast_fwd_descriptor_t *fwdDescriptor)
{
	int ret=-1;
	uint32 devModuleIndex=0xFFFFFFFF;
	rtl_multicastDeviceInfo_t devInfo;
	struct rtl_multicastFwdInfo  multicastFwdInfo;
	
	if(devName==NULL)
	{
		return FAILED;
	}

	if(multicastDataInfoPtr==NULL)
	{
		return FAILED;
	}
	
	if(fwdDescriptor==NULL)
	{
		return FAILED;
	}
	
	memset(fwdDescriptor, 0, sizeof(rtl865x_mcast_fwd_descriptor_t ));
	strcpy(fwdDescriptor->netifName,devName);
	fwdDescriptor->descPortMask=0x1<<devPortNum;
	fwdDescriptor->fwdPortMask=0xFFFFFFFF;

	if(memcmp(devName, "wlan", 4) ==0)
	{
		fwdDescriptor->toCpu=1;
	}
	
	strcpy(devInfo.devName,devName);
	
	
	ret=rtl_getDeviceIgmpSnoopingModuleIndex(&devInfo, &devModuleIndex);
	if(ret!=0)
	{
		/*un-registered device, always forwarded by cpu*/
		fwdDescriptor->toCpu=1;
	}
	else
	{
		ret= rtl_getMulticastDataFwdInfo(devModuleIndex, multicastDataInfoPtr, &multicastFwdInfo);
		if(ret!=0)
		{
			/*reserved multicast*/
			fwdDescriptor->toCpu=1;
		}
		else
		{
			if(multicastFwdInfo.cpuFlag)
			{
				fwdDescriptor->toCpu=1;
			}
			fwdDescriptor->fwdPortMask=multicastFwdInfo.fwdPortMask;
		}
	}


	return SUCCESS;
}


int rtl865x_ipMulticastHardwareAccelerate(struct net_bridge *br, unsigned int brFwdPortMask,
												unsigned int srcPort,unsigned int srcVlanId, 
												unsigned int srcIpAddr, unsigned int destIpAddr)
{
	int ret;
	int fwdDescCnt;
	struct net_bridge_port *prev;
	struct net_bridge_port *p, *n;
	unsigned short port_bitmask=0;

	rtl_multicastDeviceInfo_t devInfo;
	unsigned int	devModuleIndex;
	struct rtl_multicastDataInfo multicastDataInfo;
	struct rtl_multicastFwdInfo  multicastFwdInfo;
	
	rtl865x_tblDrv_mCast_t * existMulticastEntry;
	rtl865x_mcast_fwd_descriptor_t  fwdDescriptor;
	#if 0
	pktCnt++;
	if(pktCnt%3000==0)
	{
		printk("%s:%d,srcPort is %d,srcVlanId is %d,srcIpAddr is 0x%x,destIpAddr is 0x%x\n",__FUNCTION__,__LINE__,srcPort,srcVlanId,srcIpAddr,destIpAddr);
	}
	#endif
	if(strcmp(br->dev->name,"br0")!=0)
	{
		return -1;
	}

	//printk("%s:%d,destIpAddr is 0x%x, srcIpAddr is 0x%x, srcVlanId is %d, srcPort is %d\n",__FUNCTION__,__LINE__,destIpAddr, srcIpAddr, srcVlanId, srcPort);
	existMulticastEntry=rtl865x_findMCastEntry(destIpAddr, srcIpAddr, (unsigned short)srcVlanId, (unsigned short)srcPort);
	if(existMulticastEntry!=NULL)
	{
		if (existMulticastEntry->inAsic)
		{
			/*continue to add*/
		}
		else
		{
			/*it's already in cache */
			return 0;
		}

	}

	multicastDataInfo.ipVersion=4;
	multicastDataInfo.sourceIp[0]=  srcIpAddr;
	multicastDataInfo.groupAddr[0]=  destIpAddr;

	/*add hardware multicast entry*/
	fwdDescCnt=0;
	prev=NULL;
	
	if(brFwdPortMask==0)
	{
		rtl865x_blockMulticastFlow(srcVlanId, srcPort, srcIpAddr, destIpAddr);
		return 0;
	}

	list_for_each_entry_safe(p, n, &br->port_list, list) 
	{
		port_bitmask = (1 << p->port_no);
		if (port_bitmask & brFwdPortMask)  	 
		{
			if(p->state == BR_STATE_FORWARDING)
			{
				if (prev != NULL)
				{
					rtl865x_genMCastFwdDescriptor(p->dev->name, p->port_no,&multicastDataInfo,&fwdDescriptor);

					/*to flush old descriptor chain when add first one*/
					if(fwdDescCnt==0)
					{
						//printk("%s:%d,destIpAddr is 0x%x, srcIpAddr is 0x%x, srcVlanId is %d, srcPort is %d\n",__FUNCTION__,__LINE__,destIpAddr, srcIpAddr, srcVlanId, srcPort);
						rtl865x_addMulticastEntry(destIpAddr, srcIpAddr, (unsigned short)srcVlanId, (unsigned short)srcPort, 
						&fwdDescriptor, 1, 0, 1, 0);
					}
					else
					{	
						/*incremental adding*/
						//printk("%s:%d,destIpAddr is 0x%x, srcIpAddr is 0x%x, srcVlanId is %d, srcPort is %d\n",__FUNCTION__,__LINE__,destIpAddr, srcIpAddr, srcVlanId, srcPort);
						ret=rtl865x_addMulticastEntry(destIpAddr, srcIpAddr, (unsigned short)srcVlanId, (unsigned short)srcPort, 
						&fwdDescriptor, 0, 0, 1, 0);
					}
					
					fwdDescCnt++;
					
				}
				
				prev = p;
			}
                                                                                                                      
               }
	}
	
	//printk("%s:%d,fwdDescCnt is %d\n",__FUNCTION__,__LINE__,fwdDescCnt);
	if (prev != NULL) 
	{
		rtl865x_genMCastFwdDescriptor(prev->dev->name, prev->port_no,&multicastDataInfo,&fwdDescriptor);

		/*the first one and only one forwarding device exist*/
		if(fwdDescCnt==0)
		{
			//printk("%s:%d,destIpAddr is 0x%x, srcIpAddr is 0x%x, srcVlanId is %d, srcPort is %d\n",__FUNCTION__,__LINE__,destIpAddr, srcIpAddr, srcVlanId, srcPort);
			rtl865x_addMulticastEntry(destIpAddr, srcIpAddr, (unsigned short)srcVlanId, (unsigned short)srcPort, 
						&fwdDescriptor, 1, 0, 0, 0);
		}
		else
		{
			/*the last one, incremental adding */
			//printk("%s:%d,destIpAddr is 0x%x, srcIpAddr is 0x%x, srcVlanId is %d, srcPort is %d\n",__FUNCTION__,__LINE__,destIpAddr, srcIpAddr, srcVlanId, srcPort);
			rtl865x_addMulticastEntry(destIpAddr, srcIpAddr, (unsigned short)srcVlanId, (unsigned short)srcPort, 
						&fwdDescriptor, 0, 0, 0, 0);
		}
	}
	return 0;
}

#endif






